# -*- coding: utf-8 -*-

from __future__ import unicode_literals
import six
import datetime


from driverng import (
    MessageV2,
    Worklist
)


FIELDS_CODE_NAME_MAP = {
    'HEADER': 'MES SQA-VISION',
    'FIC': 'Fields count',
    'TFN': 'TF_Name',
    'TFA': 'Address',
    'TFP': 'Phone',
    'TFF': 'Fax',
    'TFS': 'Site',
    'TFZ': 'E-Mail'
}


# questions
# ENCODING ???

# pas de default value pour le moment
# check sur le champs FIC Field Count


class InvalidEntry(Exception):
    def __init__(self, value):
        self.value = value
        self.message = value


class ParsingError(Exception):
    def __init__(self, value):
        self.value = value
        self.message = value


class ForeignFieldContainer(list):
    def __init__(self, type_segment, cls):
        self.cls = cls
        self.type_segment = type_segment
        list.__init__(self)

    def add(self, *args, **kwargs):
        kwargs.setdefault('type_segment', self.type_segment)
        segment = self.cls(*args, **kwargs)
        self.append(segment)
        return segment


class BaseField(object):
    def get_default(self):
        raise NotImplementedError

    def to_python(self, value):
        raise NotImplementedError

    def to_string(self, value):
        raise NotImplementedError


class BaseForeignField(object):
    pass


class Field(BaseField):
    __slots__ = ['code', 'default_value']
    PARAMETER_SEPARATOR = '^'

    def __init__(self, code, default=None):
        self.code = code
        self.default_value = default

    def get_default(self):
        if self.default_value:
            return self.default_value
        return None

    def to_string(self, value=None):
        # gestion des caractère spéciaux relous?
        if self.code == 'HEADER':
            return self.code
        return '{code}{sep1}{value}{sep2}'.format(
            code=self.code,
            sep1=self.PARAMETER_SEPARATOR,
            value=value or self.default_value,
            sep2=self.PARAMETER_SEPARATOR
        )

    def to_python(self, value=None):
        return value

    @staticmethod
    def parse(entry):
        splitted_entry = entry.split(Field.PARAMETER_SEPARATOR)
        return (splitted_entry[0], splitted_entry[1])


class HeaderField(Field):
    __slots__ = ['code', 'default_value', 'is_header']

    def __init__(self, code, default=None):
        self.is_header = True
        super(HeaderField, self).__init__(code, default)


class DateField(Field):
    def to_python(self, value=''):
        return datetime.datetime.strptime(value, '%m/%d/%y')

    def to_string(self, value=''):
        if isinstance(value, datetime.Datetime):
            return value.strftime('%m/%d/%y')
        return value


class TimeField(Field):
    def to_python(self, value=''):
        return datetime.datetime.strptime(value, '%H:%M').time()

    def to_string(self, value=''):
        if isinstance(value, datetime.Time):
            return value.strftime('%H:%M')
        return value


class WithoutDayDateField(Field):
    def to_python(self, value=''):
        return datetime.datetime.strptime(value, '%m/%y')

    def to_string(self, value=''):
        return value.strftime('%m/%y')


class ForeignField(BaseForeignField):
    # TODO => enlever les attribut non utilisé
    def __init__(self, code, type_segment, relname, one2one=False):
        self.code = code
        self.type_segment = type_segment
        self.relname = relname
        self.one2one = one2one
        self.cls = None
        super(ForeignField, self).__init__()

    def copy(self):
        # hmm pk faire en vrai?
        pass

    def set_segment(self, cls):
        self.cls = cls

    def get_default(self):
        if self.one2one:
            return None
        else:
            return ForeignFieldContainer(self.type_segment, self.cls)


class Options(object):
    def __init__(self, fields, foreign_fields):
        self.fields = fields
        self.foreign_fields = foreign_fields


class BaseSegment(type):
    def __new__(cls, name, bases, attrs):
        fields = {}
        foreign_fields = {}
        for base in bases:
            if isinstance(base, BaseSegment):
                for k, v in base._meta.fields.items():
                    fields[k] = v
                for k, v in base._meta.foreign_fields.items():
                    foreign_fields[k] = v.copy()
        fields_definition = attrs.get('FIELDS')
        if fields_definition:
            for k, v in [(f.code, f) for f in fields_definition]:
                if isinstance(v, BaseField):
                    fields[k] = v
            del attrs['FIELDS']
        foreign_fields_definition = attrs.get('FOREIGN_FIELDS')
        if foreign_fields_definition:
            for k, v in [(f.code, f) for f in foreign_fields_definition]:
                if isinstance(v, BaseForeignField):
                    foreign_fields[k] = v
        new_class = super(BaseSegment, cls).__new__(cls, name, bases, attrs)
        meta = Options(fields, foreign_fields)
        new_class._meta = meta
        return new_class


@six.add_metaclass(BaseSegment)
class AbstractSegment(object):
    _meta = None

    SEPARATOR = '|'

    def __init__(self, **kwargs):
        self._parent = None
        self._code_fields_with_value = []
        self._unretrieved_code_fields = []

        for name, field in self._meta.fields.items():
            setattr(self, name, kwargs.get(name, field.get_default()))

        for name, field in self._meta.foreign_fields.items():
            # instancie le container du foreign
            setattr(self, name, field.get_default())

        super(AbstractSegment, self).__init__()

    def get(self, code):
        value = getattr(self, code)
        if value:
            return getattr(self, code).get('python')
        return value

    @staticmethod
    def attach_segment(segment, parent):
        if parent is None:
            return False

        for name, field in parent._meta.foreign_fields.items():
            if (
                field.type_segment == segment.type_segment and
                segment.type_segment != parent.type_segment
            ):
                segment._parent = parent
                setattr(segment, field.relname, parent)
                if field.one2one:
                    setattr(parent, name, segment)
                else:
                    getattr(parent, name).append(segment)
                return True

        # on a pas trouvé sur le parent direct => on remonte d'un niveau
        return AbstractSegment.attach_segment(segment, parent._parent)

    def pre_parse_entry(self, unparsed_entry):
        try:
            splitted_entry = unparsed_entry.split('^')
            if len(splitted_entry) == 1:
                code, value = None, splitted_entry[0]
            else:
                code, value, _ = splitted_entry
            return code, value
        except ValueError:
            raise InvalidEntry(unparsed_entry)

    def populate(self, unparsed_entry_list):
        for unparsed_entry in unparsed_entry_list:
            code, value = self.pre_parse_entry(unparsed_entry)
            if not code:
                code = 'HEADER'
            field = self._meta.fields.get(code)
            if not field:
                raise ParsingError('could not field field with code %s' % (code,))
            field_value = {
                'python': field.to_python(value),
                'string': value
            }
            setattr(self, code, field_value)
            self._unretrieved_code_fields.append(code)
            if not hasattr(field, 'is_header'):
                self._code_fields_with_value.append(code)

    def to_string(self):
        splitted_data = []
        for code, field in self._meta.fields.items():
            trame_value = field.to_string(getattr(self, code))
            splitted_data.append(trame_value)

        return self.SEPARATOR.join(splitted_data)

    def field_iterator(self, only_unretrieved=False, string_value=False):
        if string_value:
            sub_key = 'string'
        else:
            sub_key = 'python'
        for code in self._code_fields_with_value:
            if only_unretrieved:
                if code in self._unretrieved_code_fields:
                    value = getattr(self, code)
                else:
                    continue
            else:
                value = getattr(self, code)
            if value:
                value = value.get(sub_key)
            yield code, value

    def retrieve(self, code, string_value=False):
        '''retourne la valeur de l'attribut 'code' et le marque comme 'retrieved' sur le segment.
        arg: string_value: retourne la valeur au format 'string' au lieu du format 'python'
        '''
        try:
            code_position = self._unretrieved_code_fields.index(code)
            self._unretrieved_code_fields.pop(code_position)
        except ValueError:
            # le code a deja été retrieve une fois
            pass
        value = getattr(self, code)
        if not value:
            return value
        if string_value:
            return getattr(self, code).get('string')
        return getattr(self, code).get('python')


class SegmentH(AbstractSegment):
    FIELDS = [
        HeaderField('HEADER'),
        HeaderField('FIC'),
        Field('TFN'),
        Field('TFA'),
        Field('TFP'),
        Field('TFF'),
        Field('TFS'),
        Field('TFS'),
        Field('TFZ')
    ]

    FOREIGN_FIELDS = [
        ForeignField('patients', 'P', 'header')
    ]

    type_segment = 'H'


class SegmentP(AbstractSegment):
    FIELDS = [
        Field('PID'),
        Field('RTY'),
        HeaderField('FIC'),
        Field('PFN'),
        Field('PLN'),
        DateField('PBD')
    ]

    FOREIGN_FIELDS = [
        ForeignField('orders', 'O', 'patient')
    ]

    type_segment = 'P'


class SegmentO(AbstractSegment):
    CONTROL_TYPE_MAP = {
        '1': 'latex_beads',
        '2': 'stabilized_sperm',
        '3': 'proficiency',
    }

    FIELDS = [
        HeaderField('SN#'),
        Field('SID'),
        HeaderField('FIC'),
        Field('PVTOTIMM'),
        Field('PVTOTMOTIMM'),
        Field('PVTOTSPE'),
        Field('ABNOR'),
        Field('AB1'),
        Field('AB2'),
        Field('AB3'),
        Field('AB3'),
        Field('AB5'),
        Field('AB6'),
        Field('AB7'),
        Field('AB8'),
        Field('AB9'),
        Field('ABS'),
        Field('AGGL'),
        Field('APPR'),
        DateField('COLLDATE'),
        TimeField('COLLTIME'),
        Field('CLOC'),
        Field('COMM'),
        Field('COMSAM'),
        Field('STAND'),
        Field('CONC'),
        Field('MORPHINDEX'),
        Field('ROUNCELL'),
        Field('DIFF'),
        Field('DIRIB2'),
        Field('DIRIB1'),
        Field('DIRMAR2'),
        Field('DIRMAR1'),
        Field('FVOL'),
        Field('FRUC'),
        Field('FSC'),
        Field('FUNCSPE'),
        Field('AGLUC'),
        Field('IMMOT5'),
        Field('PVIMM'),
        Field('IMMOT'),
        Field('IVOL'),
        Field('LES'),
        Field('LIQ'),
        Field('LONGID'),
        Field('LONGS'),
        Field('MEDI'),
        Field('MORPHSPE'),
        Field('MOTSPE'),
        Field('PVMOT'),
        Field('MSC'),
        Field('MOT'),
        Field('NPMOT'),
        Field('NPMOT5'),
        Field('NLMORPH'),
        Field('OPT1'),
        Field('OPT2'),
        Field('OIN1'),
        Field('OIN2'),
        Field('OIN3'),
        Field('OIN4'),
        Field('PEROX'),
        Field('PH'),
        Field('AB10'),
        Field('PMOTSPE'),
        Field('PMSC'),
        Field('PMSCA'),
        Field('PMSCB'),
        Field('PMOT'),
        Field('PMOTA'),
        Field('RECEDATE'),
        Field('RECETIME'),
        Field('TESTM'),
        Field('SEMTRE'),
        Field('PMOTB'),
        Field('TSPE'),
        Field('SMI'),
        DateField('ADT'),
        TimeField('ATM'),
        Field('TRB'),
        Field('SAMTY'),
        Field('MOT5'),
        Field('PVTOTMOT'),
        Field('VELOC'),
        Field('VISC'),
        Field('VITD'),
        Field('VIT'),
        Field('VOL'),
        Field('WBCCONC'),
        Field('ZINC'),
        Field('CTM'),
        Field('CDT'),
        Field('CLEV'),
        Field('CTY'),
        Field('LOT'),
        WithoutDayDateField('EXD'),
        Field('TAR'),
        Field('RAN'),
        Field('CTSC'),
        Field('RST'),
        Field('COA'),
        Field('CMSC'),
        Field('STY')
    ]

    type_segment = 'O'


class Driver(MessageV2):
    SPEC = {
        'code': 'sqavision',
        'label': 'SQA-VISION'
    }
    SEGMENT_MAP = {
        'H': SegmentH,
        'P': SegmentP,
        'O': SegmentO
    }

    OPTIONS = [
        MessageV2.BooleanOption(
            'sqavision_use_SID_instead_of_PID',
            label=u'utiliser le code echantillon renseigné dans le champs SID au lieu du champs PID',
            default=False
        )
    ]

    TOP_LEVEL_SEGMENT_TYPE = 'H'

    TEST_RECORD_TYPE = '0'
    CONTROL_RECORD_TYPE = '1'

    def __init__(self, *args, **kwargs):
        self.sqavision_use_SID_instead_of_PID = kwargs.pop('sqavision_use_SID_instead_of_PID', None)
        super(Driver, self).__init__(*args, **kwargs)

    def add_segment(self, segment, message, last_segment):
        if segment.type_segment not in self.SEGMENT_MAP.keys():
            raise ParsingError()

        if message is None:
            # premier message => doit être de type H
            if not segment.type_segment == self.TOP_LEVEL_SEGMENT_TYPE:
                raise ParsingError()

            message = segment
        else:
            # on cherche le parent (le bon!) sur lequel attacher le segment
            if not AbstractSegment.attach_segment(segment, last_segment):
                raise ParsingError(u'Le segment %s n\'est pas autorisé ici' % segment.type_segment)
        last_segment = segment

        return message, last_segment

    def create(self, **kwargs):
        return SegmentH(**kwargs)

    def parse(self, data):
        lines = [l for l in data.splitlines() if l]

        message, last_segment = None, None

        i = 0
        while i < len(lines):
            line = lines[i]
            try:
                segment_klass = self.SEGMENT_MAP[line[0]]
            except KeyError:
                raise ParsingError('Le segment %s n\'est pas autorisé dans ce protole' % line[0])
            line_data = line.split(segment_klass.SEPARATOR)
            segment = segment_klass()
            segment.populate(line_data[1:])

            message, last_segment = self.add_segment(
                segment, message, last_segment)
            i += 1
        return message

    def decode_message(self, message):
        worklist = Worklist(context='result')

        worklist.sender_name = message.retrieve('TFN')

        for patient in message.patients:
            for order in patient.orders:
                if self.sqavision_use_SID_instead_of_PID:
                    sample_id = order.retrieve('SID')
                else:
                    sample_id = patient.retrieve('PID')
                sample = worklist.samples.add(
                    patient_id=patient.get('PID'),
                    sample_id=sample_id)
                if patient.get('RTY') == self.TEST_RECORD_TYPE:
                    # c'est un resultat de test
                    sample.type = 'patient'
                    sample.first_name = patient.get('PFN')
                    sample.last_name = patient.get('PLN')
                    sample.birthdate = patient.get('PBD')
                    for code, value in order.field_iterator(only_unretrieved=True, string_value=True):
                        sample.results.add(
                            code=code,
                            value=value)
                elif patient.get('RTY') == self.CONTROL_RECORD_TYPE:
                    # c'est un resultat de QC
                    sample.type = 'control'
                    sample.control_lot = order.retrieve('LOT')
                    sample.control_level = order.retrieve('CLEV')

                    if order.get('CDT') and order.get('CTM'):
                        control_date = order.get('CDT').replace(hour=order.get('CTM').hour, minute=order.get('CTM').minute)
                    else:
                        control_date = None
                    order.retrieve('CDT')
                    order.retrieve('CTM')
                    key = order.retrieve('CTY')
                    result = sample.results.add(
                        code=order.CONTROL_TYPE_MAP[key],
                        date=control_date,
                        value=order.retrieve('CTSC'),
                    )
                    # corrective action
                    result.flags.append(order.retrieve('COA'))
                    for code, value in order.field_iterator(only_unretrieved=True, string_value=True):
                        sample.results.add(
                            code=code,
                            value=value)

                else:
                    raise ParsingError('Type de Record inconue %s' % (patient.get('RTY'),))
        return worklist
