# -*- coding: utf-8 -*-
from __future__ import unicode_literals


from driverng.utils import (
    ACK,
    NACK,
    STX,
    byte2log,
)
from driverng.protocol.base import BaseProtocol
from driverng.transport.base import TransportTimedOutException


class TrameIndexList(list):
    def __init__(self, logger=None):
        list.__init__(self)
        self.cursor = 7
        self.extend(range(8))

    def validate_and_shift(self, value):
        value = int(value)
        if (self[self.cursor] + 1) % 8 == value:
            self.cursor = value
            return True
        return False


class SQAVisionProtocol(BaseProtocol):
    def __init__(self, *args, **kwargs):
        super(SQAVisionProtocol, self).__init__(**kwargs)
        self.trame_index_manager = TrameIndexList()

    def send_trames(self, trames):
        pass

    def build_trames(self, message):
        pass

    def recv_trames(self):
        '''Recoit une liste de trames sur la connexion:
        format de la trame : <STX><index><payload><ETX><CHECKSUM_char1><CHECKSUM_char2><CR><LF>

        <- DATA 1
        -> <ack>
        <- DATA 2
        -> <ack>
        <- DATA 3
        -> <nack>
        <- DATA 3
        -> <ack>

        returns False lors d'une erreur
        returns la liste des messages
        '''

        trames = b''
        message_has_started = False

        while True:
            try:
                # <STX>...
                stx = self.read(timeout=5.0)
            except TransportTimedOutException:
                # Timeout atteint, aucune trame en attente, le message est terminé.
                break
            if not stx:
                return False
            if stx[0] != STX:
                self.logger.warning('<stx> expected in %s', byte2log(stx))
                return False
            # ...<index><payload><ETX><CHECKSUM_char1><CHECKSUM_char2><CR><LF>
            data = self.readline(timeout=15)
            trame_index = data[:1]
            if not self.trame_index_manager.validate_and_shift(trame_index):
                self.logger.warning('invalid trame index for index %s', byte2log(trame_index))
                return False
            trame_header = data[1]
            if trame_header == 'H' and message_has_started:
                self.logger.info('message was ended on last trame')
                self.write(NACK)
                break
            if trame_header == 'H':
                message_has_started = True
            if self.check_trame(data):
                self.write(ACK)
                trames += self.get_payload(data) + b'\n'
            else:
                self.write(NACK)

        messages = b'\n'.join(trames.splitlines())
        return messages

    def check_trame(self, trame):
        try:
            checksum = int(trame[-4:-2], 16)
            checksum_expected = self.checksum(trame[:-4])
        except ValueError:
            self.logger.warning(
                'Missing checksum for %s',
                byte2log(trame),
            )
            return False
        if checksum != checksum_expected:
            self.logger.warning(
                'Bad checksum for %s: (%02X != %02X)',
                byte2log(trame),
                checksum,
                checksum_expected)
            return False
        return True

    def get_payload(self, trame):
        # <payload><ETX><CHECKSUM_char1><CHECKSUM_char2><CR><LF>
        payload = trame[1:-5]
        return payload

    @staticmethod
    def checksum(data):
        return (sum(bytearray(data)) & 0xFF) % 0x100
